from ..storage import Storage


class NaiveMemory(Storage):
    """
    Storage for storing data with naive approach. It uses only memory
    and will wipe some portion of stored data when number of keys
    will react certain amount.

    It's not persistent and not safe. It's here for tiny projects
    and debugging.
    """

    def __init__(self, max_size=2000000):
        self._storage = {}
        self.max_size = max_size

    async def save(self, name, value):
        if len(self._storage) >= self.max_size:
            keys = list(self._storage.keys())

            for k in keys[:self.max_size // 2]:
                self._storage.pop(k)

        self._storage[name] = value

    async def load(self, name, default=None):
        return self._storage.get(name, default)



import importlib.util
import os
import re

from .logger import logger
from .plugin import Plugin


def import_module(name, path):
    """
    Import module from specified path with specified name.

    :param name: module's name
    :param path: path to module's file
    :returns: imported module
    """

    spec = importlib.util.spec_from_file_location(name, os.path.abspath(path))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def load_plugins_from_file(path, verbose=False):
    """
    Load plugins from specified path. Plugins can be in module-level
    variable "plugin" and in module-level variable "plugins" (with list of
    plugins).

    :param path: path to file with module
    :returns: loaded module or None if no plugin found
    """

    mod = import_module(path, path)

    plugins = []

    for pl in [getattr(mod, "plugin", None), *getattr(mod, "plugins", ())]:
        if isinstance(pl, Plugin):
            plugins.append(pl)

    if len(plugins) > 1:
        logger.info('Loaded %d plugins from "%s"', len(plugins), path)
    elif len(plugins) == 1:
        logger.info('Loaded plugin from "%s"', path)
    elif verbose:
        logger.warning('No plugins found in "%s"', path)

    return plugins


def load_plugins(folder, verbose=False):
    """
    Import all plugins from target folder recursively.

    :param folder: path to target folder
    :returns: list of loaded plugins
    """

    plugins_list = []

    for name in os.listdir(folder):
        path = os.path.join(folder, name)

        if os.path.isdir(path):
            plugins_list += load_plugins(path, verbose=verbose)

        elif re.match(r"^[^_].*\.py$", name):
            plugins_list += load_plugins_from_file(path, verbose=verbose)

    return plugins_list



from types import GeneratorType
from ..backend import Backend
from ..update import Message, ReceiverType, UpdateType


class Debug(Backend):
    def __init__(self, messages=None, on_complete=None, save_replies=True):
        if isinstance(messages, GeneratorType):  # pragma: no cover
            self.messages = messages
            self.messages_count = -1

        else:
            self.messages = (m for m in messages)
            self.messages_count = len(messages)

        self.save_replies = save_replies
        self.answers = {}
        self.answers_count = 0

        self.responses = []

        self.on_complete = on_complete

    def _make_update(self, data):
        return Message(
            raw=None,
            type=UpdateType.MSG,
            text=data[0],
            attachments=(),
            sender_id=data[1],
            receiver_id=0,
            receiver_type=ReceiverType.SOLO,
            date=0,
        )

    async def perform_updates_request(self, submit_update):
        if not self.messages:
            return

        for _ in range(25):
            try:
                await submit_update(self._make_update(next(self.messages)))
            except StopIteration:
                self.messages = None
                break

    async def perform_send(self, target_id, message, attachments, kwargs):
        if self.save_replies:
            if target_id not in self.answers:
                self.answers[target_id] = []

            self.answers[target_id].append(
                (message, attachments, kwargs)
            )

        self.answers_count += 1

        if self.answers_count == self.messages_count and self.on_complete:
            self.on_complete()



import sys
import time
import select
from ..backend import Backend
from ..update import Message, ReceiverType, UpdateType


class Terminal(Backend):
    def _make_update(self, text, sender_id=1):
        return Message(
            raw=None,
            type=UpdateType.MSG,
            text=text,
            attachments=(),
            sender_id=sender_id,
            receiver_id=1,
            receiver_type=ReceiverType.SOLO,
            date=time.time(),
        )

    async def perform_updates_request(self, submit_update):
        ready_objects, _, _ = select.select([sys.stdin], [], [], 0.1)

        if ready_objects:
            message = sys.stdin.readline().strip()
            await submit_update(self._make_update(message))

    async def perform_send(self, target_id, message, attachments, kwargs):
        print(">", message)



from collections import namedtuple
from enum import Enum


class HandlerResponse(Enum):
    COMPLETE = 1
    SKIPPED = 2


Handler = namedtuple("Handler", [
    "handle", "group_state", "user_state", "priority",
])



import os.path


def get_path(root, path):
    """
    Shortcut for ``os.path.join(os.path.dirname(root), path)``.

    :param root: root path
    :param path: path to file or folder
    :returns: path to file or folder relative to root
    """

    return os.path.join(os.path.dirname(root), path)



from sortedcontainers import SortedList
from .handler import HandlerResponse as hr


class Router:
    __slots__ = ("priority",)

    def __init__(self, priority=0):
        self.priority = priority

    def _key(self, handler):
        """Key for sorting handlers."""
        return (
            handler.group_state == "*",
            handler.user_state == "*",
            -handler.priority,
        )

    def _check_update(self, update, ctx):
        """Should update be processed?"""
        return True

    def _check(self, handler, update, ctx):
        """Should this handler handle update?"""

        return (
            (handler.group_state in ("*", ctx.group_state))
            and (handler.user_state in ("*", ctx.user_state))
        )

    def merge(self, other_router):
        raise NotImplementedError

    async def handle(self, update, ctx):
        raise NotImplementedError


class ListRouter(Router):
    __slots__ = ("_handlers",)

    def __init__(self, priority=0):
        super().__init__(priority)
        self._handlers = SortedList([], key=self._key)

    def add_handler(self, handler):
        self._handlers.add(handler)

    def merge(self, other_router):
        if not isinstance(other_router, self.__class__):
            raise RuntimeError("Can't merge routers with different classes")

        self._handlers.update(other_router._handlers)

    async def handle(self, update, ctx):
        if not self._check_update(update, ctx):
            return hr.SKIPPED

        for handler in self._handlers:
            if self._check(handler, update, ctx):
                if await handler.handle(update, ctx) != hr.SKIPPED:
                    return hr.COMPLETE

        return hr.SKIPPED


class MapRouter(Router):
    __slots__ = ("_handlers",)

    def __init__(self, priority=0):
        super().__init__(priority)
        self._handlers = {}

    def _get_keys(self, update, ctx):
        """
        Returns tuple of strings for determining handlers in `_handlers`
        dictionary.
        """

        raise NotImplementedError

    def add_handler(self, handler, key):
        if key in self._handlers:
            self._handlers[key].add(handler)
        else:
            self._handlers[key] = SortedList([handler], key=self._key)

    def merge(self, other_router):
        if not isinstance(other_router, self.__class__):
            raise RuntimeError("Can't merge routers with different classes")

        for key, handlers in other_router._handlers.items():
            for handler in handlers:
                self.add_handler(handler, key)

    async def handle(self, update, ctx):
        keys = self._get_keys(update, ctx)

        for key in keys:
            if key not in self._handlers:
                continue

            for handler in self._handlers[key]:
                if self._check(handler, update, ctx):
                    if await handler.handle(update, ctx) != hr.SKIPPED:
                        return hr.COMPLETE

        return hr.SKIPPED



class Storage:
    async def initiate(self):  # pragma: no cover
        pass

    async def save(self, name, value):
        raise NotImplementedError

    async def load(self, name, default=None):
        raise NotImplementedError



try:  # pragma: no cover
    import uvloop
    uvloop.install()
except ModuleNotFoundError:  # pragma: no cover
    pass


from .kutana import Kutana
from .plugin import Plugin
from .backend import Backend
from .context import Context
from .update import Update, Message, Attachment, UpdateType
from .handler import HandlerResponse
from .exceptions import RequestException
from .loaders import load_plugins, load_plugins_from_file
from .helpers import get_path


__all__ = [
    "Kutana", "Plugin", "Backend", "Update", "Message", "Attachment",
    "UpdateType", "HandlerResponse", "RequestException", "Context",
    "load_plugins", "load_plugins_from_file", "get_path",
]



import re
from .router import MapRouter, ListRouter
from .update import UpdateType


class CommandsRouter(MapRouter):
    __slots__ = ("_cache",)

    def __init__(self, priority=6):
        """Base priority is 6."""
        super().__init__(priority=priority)
        self._cache = None

    def _populate_cache(self, ctx):
        commands = list(self._handlers.keys())
        prefixes = ctx.config["prefixes"]

        self._cache = re.compile(
            pattern=r"({prefix})({command})($|\s.*)".format(
                prefix="|".join(re.escape(p) for p in prefixes),
                command="|".join(re.escape(c) for c in commands),
            ),
            flags=re.IGNORECASE,
        )

    def add_handler(self, handler, key):
        return super().add_handler(handler, key.lower())

    def _get_keys(self, update, ctx):
        if update.type != UpdateType.MSG:
            return ()

        if self._cache is None:
            self._populate_cache(ctx)

        match = self._cache.match(update.text)

        if match is None:
            return ()

        ctx.prefix = match.group(1)
        ctx.command = match.group(2)
        ctx.body = (match.group(3) or "").strip()
        ctx.match = match

        return (ctx.command.lower(),)


class AttachmentsRouter(MapRouter):
    __slots__ = ()

    def __init__(self, priority=3):
        """Base priority is 3."""
        super().__init__(priority=priority)

    def _get_keys(self, update, ctx):
        if update.type != UpdateType.MSG:
            return ()

        return tuple(a.type for a in update.attachments)


class AnyMessageRouter(ListRouter):
    __slots__ = ()

    def __init__(self, priority=9):
        """Base priority is 9."""
        super().__init__(priority=priority)

    def _check_update(self, update, ctx):
        return update.type == UpdateType.MSG and update.text.strip()


class AnyUnprocessedMessageRouter(AnyMessageRouter):
    __slots__ = ()

    def __init__(self, priority=-3):
        """Base priority is -3."""
        super().__init__(priority=priority)



import re
from .handler import Handler, HandlerResponse
from .update import UpdateType
from .backends.vkontakte import VkontaktePluginExtension
from .routers import (
    CommandsRouter, AttachmentsRouter, ListRouter, AnyMessageRouter,
    AnyUnprocessedMessageRouter,
)


class Plugin:
    """
    Plugin provides functionality to the application.

    :param name: Name of the plugin
    :param description: Description of the plugin
    """

    def __init__(self, name, description=""):
        self.name = name
        self.description = description

        self._routers = []

        self._on_start = None
        self._on_before = None
        self._on_after = None
        self._on_exception = None
        self._on_shutdown = None

    def _get_or_add_router(self, router):
        for r in self._routers:
            if isinstance(r, router):
                return r

        r = router()
        self._routers.append(r)
        return r

    def on_start(self):
        """
        Decorator for registering coroutine to be called when application
        starts. It will be passed an application as an argument.
        """
        def decorator(func):
            if self._on_start is not None:
                raise RuntimeError(f"Hook 'on_start' already set")
            self._on_start = func
            return func
        return decorator

    def on_before(self):
        """
        Decorator for registering coroutine to be called before an update
        will be processed. It will be passed the update and it's context as
        arguments.
        """
        def decorator(func):
            if self._on_before is not None:
                raise RuntimeError(f"Hook 'on_before' already set")
            self._on_before = func
            return func
        return decorator

    def on_after(self):
        """
        Decorator for registering coroutine to be called after an update
        was processed. It will be passed the update, it's context and
        result of processing (:class:`kutana.handler.HandlerResponse`) as
        arguments.
        """
        def decorator(func):
            if self._on_after is not None:
                raise RuntimeError(f"Hook 'on_after' already set")
            self._on_after = func
            return func
        return decorator

    def on_exception(self):
        """
        Decorator for registering coroutine to be called when exception was
        raised while processing an update. It will be passed the update,
        it's context and raised exception as arguments.
        """
        def decorator(func):
            if self._on_exception is not None:
                raise RuntimeError(f"Hook 'on_exception' already set")
            self._on_exception = func
            return func
        return decorator

    def on_shutdown(self):
        """
        Decorator for registering coroutine to be called when application
        will be stopped. It will be passed application as an argument.
        """
        def decorator(func):
            if self._on_shutdown is not None:
                raise RuntimeError(f"Hook 'on_shutdown' already set")
            self._on_shutdown = func
            return func
        return decorator

    # Registrators for updates with specific conditions
    @property
    def vk(self) -> VkontaktePluginExtension:
        return VkontaktePluginExtension(self)

    def on_commands(
        self,
        commands,
        group_state="*",
        user_state="*",
        priority=0,
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message, starts with prefix and one of
        provided commands.

        If case_insensitive is True, commands will be processed with ignored
        case.

        Context is automatically populated with following values:

        - prefix
        - command
        - body
        - match

        You can specify group_state and/or user_state to control
        which state must have groups and users attempting to
        access this coroutine.

        Priority is a value that used to determine order in which
        coroutines are executed. If coroutine returns anythign but
        :class:`kutana.handler.HandlerResponse` SKIPPED, other
        coroutines are not executed further.
        """

        if not isinstance(commands, (list, tuple)):
            raise ValueError("`commands` argument must be list or tuple")

        def decorator(func):
            router = self._get_or_add_router(CommandsRouter)
            for command in commands:
                router.add_handler(
                    Handler(func, group_state, user_state, priority),
                    command,
                )
            return func

        return decorator

    def on_attachments(
        self,
        types,
        group_state="*",
        user_state="*",
        priority=0,
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message and have at least one
        attachment with one of specified types.

        You can specify group_state and/or user_state to control
        which state must have groups and users attempting to
        access this coroutine.

        Priority is a value that used to determine order in which
        coroutines are executed. If coroutine returns anythign but
        :class:`kutana.handler.HandlerResponse` SKIPPED, other
        coroutines are not executed further.
        """

        def decorator(func):
            router = self._get_or_add_router(AttachmentsRouter)
            for _type in types:
                router.add_handler(
                    Handler(func, group_state, user_state, priority),
                    _type,
                )
            return func

        return decorator

    def on_any_message(self, group_state="*", user_state="*", priority=0):
        """
        Decorator for registering coroutine to be called when
        incoming update is message. This will always be called.
        If you want to catch all unprocessed message, you should
        use :meth:`kutana.plugin.Plugin.on_any_unprocessed_message`.

        You can specify group_state and/or user_state to control
        which state must have groups and users attempting to
        access this coroutine.

        Priority is a value that used to determine order in which
        coroutines are executed. If coroutine returns anythign but
        :class:`kutana.handler.HandlerResponse` SKIPPED, other
        coroutines are not executed further.
        """

        def decorator(func):
            router = self._get_or_add_router(AnyMessageRouter)
            router.add_handler(
                Handler(func, group_state, user_state, priority),
            )
            return func

        return decorator

    def on_any_unprocessed_message(
        self,
        group_state="*",
        user_state="*",
        priority=0,
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message. This will be called if no other
        coroutine processed update.

        You can specify group_state and/or user_state to control
        which state must have groups and users attempting to
        access this coroutine.

        Priority is a value that used to determine order in which
        coroutines are executed. If coroutine returns anythign but
        :class:`kutana.handler.HandlerResponse` SKIPPED, other
        coroutines are not executed further.
        """

        def decorator(func):
            router = self._get_or_add_router(AnyUnprocessedMessageRouter)
            router.add_handler(
                Handler(func, group_state, user_state, priority),
            )
            return func

        return decorator

    def on_match(self, pattern, group_state="*", user_state="*", priority=0):
        """
        Decorator for registering coroutine to be called when
        incoming update is message and it's text is matched by
        provided pattern.

        Context is automatically populated with following values:

        - match

        You can specify group_state and/or user_state to control
        which state must have groups and users attempting to
        access this coroutine.

        Priority is a value that used to determine order in which
        coroutines are executed. If coroutine returns anythign but
        :class:`kutana.handler.HandlerResponse` SKIPPED, other
        coroutines are not executed further.
        """

        def decorator(func):
            router = self._get_or_add_router(ListRouter)

            async def wrapper(update, ctx):
                if update.type != UpdateType.MSG:
                    return HandlerResponse.SKIPPED

                match = re.match(pattern, update.text)

                if not match:
                    return HandlerResponse.SKIPPED

                ctx.match = match

                return await func(update, ctx)

            router.add_handler(
                Handler(wrapper, group_state, user_state, priority),
            )

            return wrapper

        return decorator



import types
from .update import UpdateType as ut, ReceiverType as rt


class Context:
    """
    Use Context.create for instantiation of Context object. This object
    can store ane attributes provided by routers or other plugins.

    :ivar ~.app: Context's kutana application.
    :ivar ~.update: Update currently being processed.
    :ivar ~.backend: Backend who created update.
    :ivar ~.config: Application's configuration
    :ivar ~.default_target_id: Default receiver of responses
    :ivar ~.group_uid: Group's id plus backend identity
    :ivar ~.user_uid: User's id plus backend identity
    :ivar ~.group_state: State of group targeted by update
    :ivar ~.user_state: State of user targeted by update
    """

    def __init__(self, app=None, config=None, update=None, backend=None):
        self.app = app
        self.update = update
        self.backend = backend
        self.config = config

        self.default_target_id = None
        self.group_uid = None
        self.user_uid = None

        self.group_state = None
        self.user_state = None

    @staticmethod
    async def create(app=None, config=None, update=None, backend=None):
        """Create context with specified parameters."""

        ctx = Context(
            app=app,
            config=config,
            update=update,
            backend=backend,
        )

        if update.type == ut.MSG:
            backend_identity = backend.get_identity()
            storage = app.storage

            if update.receiver_type == rt.MULTI:
                ctx.default_target_id = update.receiver_id

                ctx.group_uid = f"{update.receiver_id}{backend_identity}"
                ctx.user_uid = f"{update.receiver_id}{backend_identity}"

                ctx.group_state_key = f"_sg:{ctx.group_uid}"
                ctx.user_state_key = f"_su:{ctx.user_uid}"

                ctx.group_state = await storage.load(ctx.group_state_key, "")
                ctx.user_state = await storage.load(ctx.user_state_key, "")
            else:
                ctx.default_target_id = update.sender_id

                ctx.group_uid = ""
                ctx.user_uid = f"{update.sender_id}{backend_identity}"

                ctx.group_state_key = ""
                ctx.user_state_key = f"_su:{ctx.user_uid}"

                ctx.group_state = ""
                ctx.user_state = await storage.load(ctx.user_state_key, "")

        return ctx

    def get(self, name, default=None):
        return getattr(self, name, default)

    def replace_method(self, method_name, method):
        """
        Replaces this instance's method with specified name with passed method.

        Example:

        .. code-block:: python

            async def replacement(self, message, attachments=(), **kwargs):
                print(message)
            ctx = Context()
            ctx.replace_method("reply", replacement)
            await ctx.reply("hey")  # will print "hey"

        .. note::

            Many methods are coroutine, so you should replace coroutines with
            coroutines and normal functions with functions.

        :param method_name: method's name to replace
        :param method: method to replace with
        """

        if not callable(getattr(self, method_name, None)):
            raise ValueError(f"No method with name '{method_name}'")

        setattr(self, method_name, types.MethodType(method, self))

    @staticmethod
    def split_large_text(text, length=4096):
        """
        Split text into chunks with specified length.

        :param text: text for splitting
        :param length: maximum size for one chunk
        :returns: tuple of chunks
        """

        yield text[0: length]

        for i in range(length, len(text), length):
            yield text[i: i + length]

    async def request(self, method, **kwargs):
        """Perform request to backend's API and return response."""

        return await self.backend.perform_api_call(
            method,
            kwargs,
        )

    async def reply(self, message, attachments=(), **kwargs):
        """
        Reply with specified message and attachments. Other kwargs are
        passed to the backend. Often as additional arguments to sending the
        message.

        If message is too long - it will be splitted.
        """

        return await self.send_message(
            self.default_target_id,
            message,
            attachments,
            **kwargs
        )

    async def send_message(self, target_id, message, attachments=(), **kwargs):
        """
        Send message to specified target_id with specifiend message and
        attachments. Other kwargs are passed to the backend. Often as
        additional arguments to sending the message.

        If message is too long - it will be splitted.
        """

        responses = []

        parts = tuple(self.split_large_text(message))

        for part in parts[:-1]:
            responses.append(await self.backend.perform_send(
                target_id,
                part,
                (),
                None,
            ))

        responses.append(await self.backend.perform_send(
            target_id,
            parts[-1],
            attachments,
            kwargs,
        ))

        return responses

    async def set_state(self, group_state=None, user_state=None):
        """
        Set group_state and/or user_state for current group and/or user.
        States can be represented by any string or None (if it shouldn't
        change).
        """

        if group_state is not None:
            if not isinstance(group_state, str):
                raise ValueError("'group_state' must be string")
            if self.group_uid == "":
                raise ValueError("You can't set group state in solo chat")
            await self.app.storage.save(self.group_state_key, group_state)

        if user_state is not None:
            if not isinstance(user_state, str):
                raise ValueError("'user_state' must be string")
            if self.user_uid == "":
                raise ValueError("You can't set group state in solo chat")
            await self.app.storage.save(self.user_state_key, user_state)



import logging
import logging.handlers


FORMATTER = logging.Formatter("%(asctime)s [ %(levelname)s ] %(message)s")
LEVEL = logging.INFO

handler_file = logging.handlers.RotatingFileHandler(
    "kutana.log", maxBytes=5243000, backupCount=5
)
handler_file.setLevel(LEVEL)
handler_file.setFormatter(FORMATTER)

handler_stream = logging.StreamHandler()
handler_stream.setLevel(LEVEL)
handler_stream.setFormatter(FORMATTER)

logger = logging.getLogger("kutana")
logger.setLevel(LEVEL)
logger.addHandler(handler_stream)
logger.addHandler(handler_file)


def set_logger_level(level):  # pragma: no cover
    handler_file.setLevel(level)
    handler_stream.setLevel(level)
    logger.setLevel(level)



class RequestException(Exception):
    def __init__(self, backend, request, response, error=None):
        self.backend = backend
        self.request = request
        self.response = response
        self.error = error



from random import random
import asyncio
import json
import re
import aiohttp
from ...logger import logger
from ...backend import Backend
from ...update import (
    ReceiverType, UpdateType, Update, Message, Attachment,
)
from ...exceptions import RequestException


NAIVE_CACHE = {}


class VKRequest(asyncio.Future):
    def __init__(self, method, kwargs):
        super().__init__()
        self.method = method
        self.kwargs = kwargs


class Vkontakte(Backend):
    def __init__(
        self,
        token,
        session=None,
        requests_per_second=19,
        longpoll_settings=None,
        api_version="5.103",
        api_url="https://api.vk.com"
    ):
        if not token:
            raise ValueError("No `token` specified")

        self.api_token = token
        self.api_version = api_version
        self.api_request_pause = 1 / requests_per_second

        self.session = session
        self._is_session_local = session is None

        self.group_id = None
        self.group_screen_name = None
        self.group_name = None

        self.longpoll_settings = longpoll_settings or {}
        self.longpoll_data = None

        self.requests_queue = []

        self.api_request_url = \
            api_url + f"/method/{{}}?access_token={token}&v={api_version}"

        self.longpoll_url_template = "{}?act=a_check&key={}&wait=25&ts={}"

    async def _get_response(self, method, kwargs={}):
        data = {k: v for k, v in kwargs.items() if v is not None}

        request_url = self.api_request_url.format(method)

        async with self.session.post(request_url, data=data) as response:
            return await response.json(content_type=None)

    async def raw_request(self, method, kwargs={}):
        """
        Call specified method from VKontakte api with specified kwargs
        and return response's data.

        This method raises RequestException if response contains error.
        """

        data = await self._get_response(method, kwargs)

        if not data.get("response"):
            raise RequestException(self, (method, {**kwargs}), data)

        return data["response"]

    async def _execute_loop(self, loop):
        while True:
            await asyncio.sleep(self.api_request_pause)

            requests = []

            for _ in range(25):
                if not self.requests_queue:
                    break

                requests.append(self.requests_queue.pop())

            if not requests:
                continue

            code = "return ["

            for r in requests:
                kwargs = json.dumps(r.kwargs, ensure_ascii=False)
                code += f"API.{r.method}({kwargs}),"

            code += "];"

            asyncio.ensure_future(
                self._execute_loop_perform_execute(code, requests),
                loop=loop,
            )

    async def _execute_loop_perform_execute(self, code, requests):
        response = await self._get_response("execute", {"code": code})

        result = response.get("response") or []
        errors = response.get("execute_errors") or []

        if len(result) != len(requests):
            result = [*result, *[False] * (len(requests) - len(result))]

        if len(errors) != len(requests):
            errors = [*errors, *[None] * (len(requests) - len(errors))]

        for res, req, err in zip(result, requests, errors):
            if req.done():
                continue

            if res is False or res is None:
                req.set_exception(RequestException(
                    self,
                    (req.method, req.kwargs),
                    res,
                    err,
                ))
            else:
                req.set_result(res)

    async def _request(self, method, kwargs, timeout=None):
        req = VKRequest(method, kwargs)

        self.requests_queue.append(req)

        res = await asyncio.wait_for(req, timeout=timeout)

        logger.debug("Vkontakte: %s(%s) => %s", method, kwargs, res)

        return res

    async def update_longpoll_data(self):
        longpoll = await self.raw_request(
            "groups.getLongPollServer", {"group_id": self.group_id}
        )

        self.longpoll_data = longpoll.copy()

    async def resolve_screen_name(self, screen_name):
        if screen_name in NAIVE_CACHE:
            return NAIVE_CACHE[screen_name]

        result = await self._request(
            "utils.resolveScreenName",
            {"screen_name": screen_name}
        )

        if len(NAIVE_CACHE) >= 500_000:
            NAIVE_CACHE.clear()

        NAIVE_CACHE[screen_name] = result

        return result

    async def _update_group_data(self):
        groups = await self.raw_request("groups.getById", {
            "fields": "screen_name"
        })

        self.group_id = groups[0]["id"]
        self.group_name = groups[0]["name"]
        self.group_screen_name = groups[0]["screen_name"]

    def prepare_context(self, ctx):
        ctx.resolve_screen_name = self.resolve_screen_name

    def _make_getter(self, url):
        async def getter():
            async with self.session.get(url) as response:
                return await response.read()
        return getter

    def _make_attachment(self, raw_attachment):
        t = raw_attachment["type"]
        d = raw_attachment[t]

        if "owner_id" in d and "id" in d:
            id = f"{t}{d['owner_id']}_{d['id']}"
        else:
            id = None

        if t == "photo":
            return Attachment._existing_full(
                id=id, type="image", title=d["text"], file_name=id,
                getter=self._make_getter(d["sizes"][-1]["url"]), raw=d,
            )

        elif t == "doc":
            return Attachment._existing_full(
                id=id, type="doc", title=d.get("title", ""), file_name=id,
                getter=self._make_getter(d["url"]), raw=d,
            )

        elif t == "audio_message":
            return Attachment._existing_full(
                id=id, type="voice", title="", file_name=id,
                getter=self._make_getter(d["link_mp3"]), raw=d,
            )

        elif t == "sticker":
            id = f"{d['sticker_id']}"
            url = list(sorted(d["images"], key=lambda i: i["width"]))[-1]

            return Attachment._existing_full(
                id=id, type="sticker", title="", file_name=id,
                getter=self._make_getter(url), raw=d,
            )

        elif t == "audio":
            return Attachment._existing_full(
                id=id, type="audio", title=d["artist"] + " - " + d["title"],
                file_name=id, getter=self._make_getter(d["url"]), raw=d,
            )

        elif t == "video":
            return Attachment._existing_full(
                id=id, type="video", title=d["title"], file_name=id,
                getter=None, raw=d,
            )

        else:
            return Attachment._existing_full(
                id=id, type=t, title=None, file_name=None, getter=None,
                raw=d,
            )

    def _make_update(self, raw_update):
        raw_update_type = raw_update["type"]
        raw_update_object = raw_update["object"]

        if raw_update_type != "message_new":
            return Update(raw_update, UpdateType.UPD)

        raw_message = raw_update_object["message"]

        message_text = raw_message["text"]

        if raw_message["peer_id"] > 2000000000:
            receiver_type = ReceiverType.MULTI

            def sub(match):
                possible_ids = (
                    f"club{self.group_id}",
                    f"public{self.group_id}",
                    self.group_screen_name,
                )

                if match.group(1) in possible_ids:
                    return ""
                else:
                    return match.group(0)

            message_text = re.sub(r"\[(.+?)\|.+?\]", sub, message_text)
            message_text = message_text.lstrip()
        else:
            receiver_type = ReceiverType.SOLO

        attachments = []

        for attachment in raw_message.get("attachments") or ():
            attachments.append(self._make_attachment(attachment))

        return Message(
            raw=raw_update,
            type=UpdateType.MSG,
            text=message_text,
            attachments=attachments,
            sender_id=raw_message["from_id"],
            receiver_id=raw_message["peer_id"],
            receiver_type=receiver_type,
            date=raw_message["date"],
        )

    async def _upload_file_to_vk(self, upload_url, data):
        async with self.session.post(upload_url, data=data) as resp:
            return await resp.json(content_type=None)

    async def upload_attachment(self, attachment, peer_id=None):
        """
        Upload specified attachment to VKontakte with specified peer_id and
        return newly uploaded attachment.

        This method doesn't change passed attachments.
        """

        attachment_type = attachment.type

        if attachment_type == "voice":
            attachment_type = "doc"
            doctype = "audio_message"
        elif attachment_type == "graffiti":
            attachment_type = "doc"
            doctype = "graffiti"
        else:
            doctype = "doc"

        if attachment_type == "doc":
            if peer_id and doctype != "graffiti":
                upload_data = await self._request(
                    "docs.getMessagesUploadServer",
                    {"peer_id": peer_id, "type": doctype},
                )
            else:
                upload_data = await self._request(
                    "docs.getWallUploadServer",
                    {"group_id": self.group_id, "type": doctype},
                )

            data = aiohttp.FormData()
            data.add_field(
                "file", attachment.file, filename=attachment.file_name,
            )

            upload_result = await self._upload_file_to_vk(
                upload_data["upload_url"], data
            )

            attachment = await self._request(
                "docs.save", upload_result
            )

            return self._make_attachment(attachment)

        if attachment_type == "image":
            upload_data = await self._request(
                "photos.getMessagesUploadServer", {"peer_id": peer_id}
            )

            data = aiohttp.FormData()
            data.add_field(
                "photo", attachment.file, filename=attachment.file_name
            )

            upload_result = await self._upload_file_to_vk(
                upload_data["upload_url"], data
            )

            try:
                attachments = await self._request(
                    "photos.saveMessagesPhoto", upload_result
                )
            except RequestException as e:
                if not peer_id or not e.error or e.error["error_code"] != 1:
                    raise

                return await self.upload_attachment(attachment, peer_id=None)

            return self._make_attachment({
                "type": "photo",
                "photo": attachments[0],
            })

        raise ValueError(f"Can't upload attachment '{attachment_type}'")

    async def _get_updates(self):
        longpoll_url = self.longpoll_url_template.format(
            self.longpoll_data["server"],
            self.longpoll_data["key"],
            self.longpoll_data["ts"],
        )

        try:
            async with self.session.post(longpoll_url) as resp:
                return await resp.json(content_type=None)

        except (json.JSONDecodeError, aiohttp.ClientError):
            return None

        except asyncio.CancelledError:
            raise

        except Exception:
            logger.exception("Exceptions while gettings updates (Vkontakte)")
            await asyncio.sleep(1)
            return None

    async def perform_updates_request(self, submit_update):
        response = await self._get_updates()

        if response is None:
            return

        if "ts" in response:
            self.longpoll_data["ts"] = response["ts"]

        if "failed" in response:
            if response["failed"] in (2, 3):
                await self.update_longpoll_data()
            return

        for update_data in response["updates"]:
            if "type" not in update_data or "object" not in update_data:
                continue

            if update_data["type"] == "group_change_settings":
                changes = update_data["object"]["changes"]

                screen_name = changes.get("screen_name")
                if screen_name:
                    self.group_screen_name = screen_name["new_value"]

                title = changes.get("title")
                if title:
                    self.group_name = title["new_value"]

            update = self._make_update(update_data)

            await submit_update(update)

    async def perform_send(self, target_id, message, attachments, kwargs):
        # Form proper arguments
        true_kwargs = {"message": message, "peer_id": target_id}
        true_kwargs.update(kwargs)

        if "random_id" not in kwargs:
            true_kwargs["random_id"] = int(random() * 4294967296) - 2147483648

        # Form proper attachments
        true_attachments = ""

        if isinstance(attachments, (int, str, Attachment)):
            attachments = (attachments,)

        for a in attachments:
            if isinstance(a, Attachment):
                if a.type == "sticker":
                    true_kwargs["sticker_id"] = a.id
                    continue

                if not a.uploaded:
                    a = await self.upload_attachment(a, peer_id=target_id)

                if not a.id:
                    raise ValueError("Attachment has no ID")

                true_attachments += f"{a.id}"

                access_key = (a.raw or {}).get("access_key")
                if access_key:
                    true_attachments += "_" + access_key

            else:
                true_attachments += str(a)

            true_attachments += ","

        # Add attachments to arguments
        if true_attachments[:-1]:
            true_kwargs["attachment"] = true_attachments[:-1]

        return await self._request("messages.send", true_kwargs)

    async def perform_api_call(self, method, kwargs):
        return await self._request(method, kwargs)

    async def on_start(self, app):
        if not self.session:
            self.session = aiohttp.ClientSession()

        await self._update_group_data()

        longpoll_settings = dict(
            message_new=1, message_reply=0, message_allow=0,
            message_deny=0, message_edit=0, photo_new=0, audio_new=0,
            video_new=0, wall_reply_new=0, wall_reply_edit=0,
            wall_reply_delete=0, wall_reply_restore=0, wall_post_new=0,
            wall_repost=0, board_post_new=0, board_post_edit=0,
            board_post_restore=0, board_post_delete=0, photo_comment_new=0,
            photo_comment_edit=0, photo_comment_delete=0,
            photo_comment_restore=0, video_comment_new=0,
            video_comment_edit=0, video_comment_delete=0,
            video_comment_restore=0, market_comment_new=0,
            market_comment_edit=0, market_comment_delete=0,
            market_comment_restore=0, poll_vote_new=0, group_join=0,
            group_leave=0, group_change_settings=1, group_change_photo=0,
            group_officers_edit=0, user_block=0, user_unblock=0
        )

        longpoll_settings.update(self.longpoll_settings)

        await self.raw_request(
            "groups.setLongPollSettings",
            {
                "group_id": self.group_id,
                "api_version": self.api_version,
                "enabled": 1,
                **longpoll_settings,
            }
        )

        await self.update_longpoll_data()

        logger.info(
            'logged in as "%s" ( https://vk.com/club%s )',
            self.group_name,
            self.group_id,
        )

        loop = app.get_loop()

        asyncio.ensure_future(
            self._execute_loop(loop),
            loop=loop
        )

    async def send_message(self, target_id, message, attachments=(), **kwargs):
        """
        Send message to specified `target_id` with text `message` and
        attachments `attachments`.

        This method will forward all excessive keyword arguments to
        sending method.
        """

        return await self.perform_send(target_id, message, attachments, kwargs)

    async def request(self, method, _timeout=None, **kwargs):
        """
        Call specified method from VKontakte api with specified
        kwargs and return response's data.

        This method respects limits.

        This method raises RequestException if response
        contains error.
        """

        return await self._request(method, kwargs, _timeout)

    async def on_shutdown(self, app):
        if self._is_session_local:
            await self.session.close()



import json
from ...routers import MapRouter
from ...update import UpdateType
from ...handler import Handler


class PayloadRouter(MapRouter):
    __slots__ = ()

    def __init__(self, priority=7):
        """Base priority is 7"""
        super().__init__(priority=priority)

    def _to_hashable(self, obj):
        if isinstance(obj, dict):
            return tuple(
                (k, self._to_hashable(v)) for k, v in sorted(obj.items())
            )
        if isinstance(obj, list):
            return tuple(self._to_hashable(o) for o in obj)
        return obj

    def _get_keys(self, update, ctx):
        backend_identity = ctx.backend.get_identity()

        if update.type != UpdateType.MSG or backend_identity != "vkontakte":
            return ()

        message = update.raw["object"]["message"]

        try:
            payload = json.loads(message.get("payload", ""))
        except json.JSONDecodeError:
            return ()

        return (self._to_hashable(payload),)

    def add_handler(self, handler, key):
        return super().add_handler(handler, self._to_hashable(key))


class VkontaktePluginExtension:
    def __init__(self, plugin):
        self.plugin = plugin

    def on_payload(
        self,
        payloads,
        group_state="*",
        user_state="*",
        priority=0
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message and payload have specified
        content.

        You can specify group_state and/or user_state to control
        which state must have groups and users attempting to
        access this coroutine.

        Priority is a value that used to determine order in which
        coroutines are executed. If coroutine returns anythign but
        :class:`kutana.handler.HandlerResponse` SKIPPED, other
        coroutines are not executed further.
        """

        def decorator(func):
            router = self.plugin._get_or_add_router(PayloadRouter)
            for payload in payloads:
                router.add_handler(
                    Handler(func, group_state, user_state, priority),
                    payload,
                )
            return func

        return decorator



import asyncio
import aiohttp
import json
from ..backend import Backend
from ..update import Message, ReceiverType, Update, UpdateType, Attachment
from ..exceptions import RequestException
from ..logger import logger


class Telegram(Backend):
    def __init__(
        self,
        token,
        messages_per_second=29,
        session=None,
        proxy=None,
        api_url="https://api.telegram.org"
    ):
        if not token:
            raise ValueError("No `token` specified")

        self.offset = 0

        self.proxy = proxy
        self.session = session
        self._is_session_local = session is None

        self.api_token = token
        self.api_messages_pause = 1 / messages_per_second
        self.api_messages_lock = None

        self.api_url = f"https://api.telegram.org/bot{token}/{{}}"
        self.file_url = f"https://api.telegram.org/file/bot{token}/{{}}"

    async def _request(self, method, kwargs={}):
        if not self.session:
            self.session = aiohttp.ClientSession()

        data = {k: v for k, v in kwargs.items() if v is not None}

        url = self.api_url.format(method)

        async with self.session.post(url, proxy=self.proxy, data=data) as resp:
            data = await resp.json(content_type=None)

            if not data.get("ok"):
                raise RequestException(self, (method, {**kwargs}), data)

        res = data["result"]

        logger.debug("Telegram: %s(%s) => %s", method, kwargs, res)

        return res

    async def _request_file(self, file_id):
        file = await self._request("getFile", {"file_id": file_id})

        url = self.file_url.format(file["file_path"])

        async with self.session.get(url, proxy=self.proxy) as resp:
            return await resp.read()

    def _make_getter(self, file_id):
        async def getter():
            return await self._request_file(file_id)
        return getter

    def _make_attachment(self, raw_attachment, raw_attachment_type):
        t = raw_attachment_type
        d = raw_attachment

        if "file_id" in d:
            id = d["file_id"]
        else:
            id = None

        if t == "photo":
            photo = list(sorted(d, key=lambda p: p["width"]))[-1]
            id = photo["file_id"]
            return Attachment._existing_full(
                id=id, type="image", title="", file_name=id,
                getter=self._make_getter(id), raw=d,
            )

        elif t == "audio":
            title = d.get("performer", "") + " - " + d.get("title", "")
            return Attachment._existing_full(
                id=id, type="audio", title=title,
                file_name=id, getter=self._make_getter(id), raw=d,
            )

        elif t == "document":
            return Attachment._existing_full(
                id=id, type="doc", title="",
                file_name=d.get("file_name", ""),
                getter=self._make_getter(id), raw=d,
            )

        elif t == "sticker":
            return Attachment._existing_full(
                id=id, type="sticker", title="", file_name=id,
                getter=self._make_getter(id), raw=d,
            )

        elif t == "voice":
            return Attachment._existing_full(
                id=id, type="voice", title="", file_name=id,
                getter=self._make_getter(id), raw=d,
            )

        elif t == "video":
            return Attachment._existing_full(
                id=id, type="video", title="", file_name=id,
                getter=self._make_getter(id), raw=d,
            )

        else:
            return Attachment._existing_full(
                id=None, type=t, title=None, file_name=None, getter=None,
                raw=d,
            )

    def _make_update(self, raw_update):
        if "message" not in raw_update:
            return Update(raw_update, UpdateType.UPD)

        attachments = []

        possible_types = (
            "audio", "voice", "photo", "video", "document", "sticker",
            "animation", "video_note", "contact", "location", "venue",
            "poll", "invoice"
        )

        for key in possible_types:
            if key in raw_update["message"]:
                attachments.append(
                    self._make_attachment(raw_update["message"][key], key)
                )

        if raw_update["message"]["chat"]["type"] == "private":
            receiver_type = ReceiverType.SOLO
        else:
            receiver_type = ReceiverType.MULTI

        return Message(
            raw=raw_update,
            type=UpdateType.MSG,
            text=raw_update["message"].get("text", ""),
            attachments=attachments,
            sender_id=raw_update["message"]["from"]["id"],
            receiver_id=raw_update["message"]["chat"]["id"],
            receiver_type=receiver_type,
            date=raw_update["message"]["date"],
        )

    async def perform_updates_request(self, submit_update):
        try:
            response = await self._request(
                "getUpdates", {"timeout": 25, "offset": self.offset}
            )
        except (json.JSONDecodeError, aiohttp.ClientError):
            return

        except asyncio.CancelledError:
            raise

        except Exception:
            logger.exception("Exceptions while gettings updates (Telegram)")
            await asyncio.sleep(1)
            return

        for update in response:
            await submit_update(self._make_update(update))
            self.offset = update["update_id"] + 1

    async def perform_send(self, target_id, message, attachments, kwargs):
        result = []

        chat_id = str(target_id)

        async with self.api_messages_lock:
            if message:
                result.append(await self._request("sendMessage", {
                    "chat_id": chat_id,
                    "text": message,
                    **kwargs,
                }))

                await asyncio.sleep(self.api_messages_pause)

            if isinstance(attachments, (int, str, Attachment)):
                attachments = (attachments,)

            new_attachments = []

            for a in attachments:
                if isinstance(a, Attachment):
                    attach_type = {
                        "doc": "document",
                        "image": "photo",
                    }.get(a.type, a.type)

                    if a.uploaded:
                        new_attachments.append((attach_type, str(a.id)))
                    else:
                        new_attachments.append((attach_type, a.file))

            for atype, acontent in new_attachments:
                if atype not in (
                    "document", "photo", "audio", "video", "voice", "sticker"
                ):
                    raise ValueError(f"Can't upload attachment '{atype}'")

                result.append(
                    await self._request(
                        "send" + atype.capitalize(),
                        {"chat_id": chat_id, atype: acontent}
                    )
                )

                await asyncio.sleep(self.api_messages_pause)

            return result

    async def perform_api_call(self, method, kwargs):
        return await self._request(method, kwargs)

    async def on_start(self, app):
        me = await self._request("getMe")

        name = me["first_name"]
        if me.get("last_name"):
            name += " " + me["last_name"]

        logger.info(
            'logged in as "%s" ( https://t.me/%s )',
            name,
            me["username"],
        )

        self.api_messages_lock = asyncio.Lock(loop=app.get_loop())

    async def send_message(self, target_id, message, attachments=(), **kwargs):
        """
        Send message to specified `target_id` with text `message` and
        attachments `attachments`.

        This method will forward all excessive keyword arguments to
        sending method.
        """

        return await self.perform_send(target_id, message, attachments, kwargs)

    async def request(self, method, **kwargs):
        """
        Call specified method from Telegram api with specified
        kwargs and return response's data.
        """

        return await self._request(method, kwargs)

    async def on_shutdown(self, app):
        if self._is_session_local:
            await self.session.close()



from sortedcontainers import SortedList
import signal
import asyncio
from .handler import HandlerResponse as hr
from .storages import NaiveMemory
from .context import Context
from .logger import logger


class Kutana:
    """
    Main class for kutana application

    :ivar ~.storage: Storage for things like states, e.t.c.
    :ivar ~.config: Application's configuration
    """

    def __init__(
        self,
        concurrent_handlers_count=3_000,
        storage=None,
        loop=None,
    ):
        self._plugins = []
        self._backends = []

        self._loop = loop or asyncio.new_event_loop()

        self._sem = asyncio.Semaphore(
            value=concurrent_handlers_count, loop=self._loop
        )

        self._routers = None

        self.storage = storage

        self.config = {
            "prefixes": (".", "/"),
        }

    def get_loop(self):
        """Return application's asyncio loop."""
        return self._loop

    def add_plugin(self, plugin):
        """Add plugin to the application."""
        if plugin in self._plugins:
            raise RuntimeError("Plugin already added")
        self._plugins.append(plugin)

    def add_plugins(self, plugins):
        """Add every plugin in passed iterable to the application."""
        for plugin in plugins:
            self.add_plugin(plugin)

    def get_plugins(self):
        """Return list of added plugins."""
        return self._plugins

    def add_backend(self, backend):
        """Add backend to the application."""
        if backend in self._backends:
            raise RuntimeError("Backend already added")
        self._backends.append(backend)

    def get_backends(self):
        return self._backends

    async def _on_start(self, queue):
        for backend in self._backends:
            await backend.on_start(self)

            async def perform_updates_request(backend):
                def submit_update(update):
                    return queue.put((update, backend))

                while True:
                    await backend.perform_updates_request(submit_update)
                    await asyncio.sleep(0)

            asyncio.ensure_future(
                perform_updates_request(backend),
                loop=self._loop
            )

        for plugin in self._plugins:
            if plugin._on_start:
                await plugin._on_start(self)

    async def _main_loop(self):
        if self.storage is None:
            self.storage = NaiveMemory()

        queue = asyncio.Queue(maxsize=32, loop=self._loop)

        await self._on_start(queue)

        while True:
            await self._sem.acquire()

            update, backend = await queue.get()

            ctx = await Context.create(
                app=self,
                config=self.config,
                update=update,
                backend=backend,
            )

            backend.prepare_context(ctx)

            task = asyncio.ensure_future(
                self._handle_update_with_logger(update, ctx),
                loop=self._loop
            )

            task.add_done_callback(lambda t: self._sem.release())

    def _init_routers(self):
        self._routers = SortedList([], key=lambda r: -r.priority)

        def _add_router(new_router):
            for router in self._routers:
                if isinstance(router, new_router.__class__):
                    router.merge(new_router)
                    return

            self._routers.add(new_router)

        for plugin in self._plugins:
            for router in plugin._routers:
                _add_router(router)

    async def _handle_update_with_logger(self, update, ctx):
        logger.debug("Processing update %s", update)

        try:
            return await self._handle_update(update, ctx)
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            logger.exception("Exception while handling the update")

            for plugin in self._plugins:
                if plugin._on_exception:
                    await plugin._on_exception(update, ctx, exc)

    async def _handle_update(self, update, ctx):
        if self._routers is None:
            self._init_routers()

        for plugin in self._plugins:
            if plugin._on_before:
                await plugin._on_before(update, ctx)

        for router in self._routers:
            if await router.handle(update, ctx) != hr.SKIPPED:
                ctx._result = hr.COMPLETE
                break
        else:
            ctx._result = hr.SKIPPED

        for plugin in self._plugins:
            if plugin._on_after:
                await plugin._on_after(update, ctx, ctx._result)

        return ctx._result

    async def _shutdown(self):
        logger.info("Gracecfully shutting application down...")

        # Clean up
        tasks = []

        for backend in self._backends:
            tasks.append(backend.on_shutdown(self))

        for plugin in self._plugins:
            if plugin._on_shutdown:
                tasks.append(plugin._on_shutdown(self))

        await asyncio.gather(*tasks, loop=self._loop, return_exceptions=True)

        # Cancel everything else
        tasks = []

        for task in asyncio.Task.all_tasks(loop=self._loop):
            if task is not asyncio.Task.current_task():
                task.cancel()
                tasks.append(task)

        await asyncio.gather(*tasks, loop=self._loop, return_exceptions=True)

        self._loop.stop()

    def run(self):
        """Start the application."""
        logger.info("Starting application...")

        for sig in (signal.SIGHUP, signal.SIGTERM, signal.SIGINT):
            def signal_handler():  # pragma: no cover
                asyncio.ensure_future(self._shutdown(), loop=self._loop)
            self._loop.add_signal_handler(sig, signal_handler)

        try:
            asyncio.ensure_future(self._main_loop(), loop=self._loop)
            self._loop.run_forever()
        finally:
            self._loop.close()
            logger.info("Stopped application")

    def stop(self):
        """Stop the application."""
        return asyncio.ensure_future(self._shutdown(), loop=self._loop)



class Backend:
    async def on_start(self, app):
        pass

    async def on_shutdown(self, app):
        pass

    def prepare_context(self, ctx):
        pass

    async def perform_updates_request(self, submit_update):
        raise NotImplementedError

    async def perform_send(self, target_id, message, attachments, kwargs):
        raise NotImplementedError

    async def perform_api_call(self, method, kwargs):
        raise NotImplementedError

    @classmethod
    def get_identity(cls):
        return cls.__name__.lower()



from collections import namedtuple
from enum import Enum


class UpdateType(Enum):
    MSG = 1
    UPD = 2


Update = namedtuple("Update", ("raw", "type"))


class ReceiverType(Enum):
    SOLO = 1
    MULTI = 2
    UNKNOWN = 3


Message = namedtuple("Message", (
    "raw", "type", "text", "attachments", "sender_id",
    "receiver_id", "receiver_type", "date",
))


ATTACHMENT_TYPES = (
    "image", "doc", "sticker", "video", "audio", "voice"
)


AttachmentData = namedtuple("AttachmentData", (
    "id", "type", "title", "file", "file_name", "file_getter",
    "raw",
))


class Attachment(AttachmentData):
    __slots__ = ()

    @classmethod
    def new(
        cls, file, file_name="image.png", type="image", title=""
    ):
        """
        Create a new attachment with provided type, file (contents) and
        file_name. Also accepts title.

        :rtype: kutana.update.Attachment
        """

        return Attachment(
            None,
            type,
            title,
            file,
            file_name,
            None,
            None,
        )

    @classmethod
    def existing(cls, id, type):
        """
        Create an attachment that was already uploaded and you have it's ID
        and type.

        :rtype: kutana.update.Attachment
        """

        return Attachment(
            id,
            type,
            None,
            None,
            None,
            None,
            "missing",
        )

    @classmethod
    def _existing_full(
        cls, id, type, title, file_name, getter, raw,
    ):
        """
        This method is used internally for creating attachments
        with concrete data.

        :rtype: kutana.update.Attachment
        """

        return Attachment(
            id,
            type,
            title,
            None,
            file_name,
            getter,
            raw,
        )

    @property
    def uploaded(self):
        return self.raw is not None

    async def get_file(self):
        """
        Download file stored in attachments if available. Raises
        ValueError if contents can't be downloaded.

        :rtype: kutana.update.Attachment
        """

        if self.file is not None:
            return self.file

        if self.file_getter is not None:
            return await self.file_getter()

        raise ValueError("No way to get file from attachment")
